"""
Builds reduced unit peptide in PII conformation.
"""

import math
import string
import array
import os
import redux_2_1

DEGRAD = math.pi/180.0
RADDEG = 180.0/math.pi
 
def CUMSUM(a):
    l = [a[0]]
    for i in xrange(1, len(a)):
        l.append(a[i] + l[-1])
    return l

def ZEROS(n, dummy):
    return [0.0]*n
 
def ARRAY(v, dummy):
    return list(v)

#print ZEROS(3, 'd')
#print ARRAY((1.0, 0, 0), 'd')

class Atom:
    def __init__(self, name, p1, d1, d2, d3):
        self.name = name
        self.p1 = p1        # first parent
        self.d1 = d1        # distance from first parent
        self.d2 = d2        # angle to p1 and p2
        self.d3 = d3        # torsion angle 

    def copy(self):
        return Atom(self.name, self.p1, self.d1, self.d2, self.d3)


class Residue:
    def __init__(self, name, atoms=None):
        self.name = name
        if not atoms is None:
            self.atoms = atoms
        else:
            self.atoms = []
    def add_atom(self, atom):
        self.atoms.append(atom)

    def __getitem__(self, i):
        return self.atoms[i]

    def __len__(self):
        return len(self.atoms)

    def copy(self):
        newRes = Residue(self.name)
        for atom in self:
            newRes.add_atom(atom.copy())
        return newRes

def int2vec(v1, dis, v2, ang, v3, tors, sin=math.sin,\
            cos=math.cos, sqrt=math.sqrt):
    """int2vec(v1, dis, v2, ang, v3, tors)
        Compute coordinates for vector v, given its 'dis'tance
        from vector v1, 'ang'le v-v1-v2 and 'tors'ion v-v1-v2-v3.'
        Both ang and tors should be in degrees.
    """

    ang = ang*DEGRAD
    sina = sin(ang)
    cosa = cos(ang)

    tors = (tors*DEGRAD)
    sint = sina * sin(tors)
    cost = sina * cos(tors)

    x1, y1, z1 = v1
    x2, y2, z2 = v2
    x3, y3, z3 = v3

    u1x = x2 - x3
    u1y = y2 - y3
    u1z = z2 - z3
    d = 1.0 / sqrt(u1x*u1x + u1y*u1y + u1z*u1z)
    u1x = u1x*d
    u1y = u1y*d
    u1z = u1z*d

    u2x = x1 - x2
    u2y = y1 - y2
    u2z = z1 - z2;
    d = 1.0 / sqrt(u2x*u2x + u2y*u2y + u2z*u2z)
    u2x = u2x*d
    u2y = u2y*d
    u2z = u2z*d

    cosine = u1x*u2x + u1y*u2y + u1z*u2z

    if (abs(cosine) < 1.0):
        sine = 1.0/sqrt(1.0 - cosine*cosine)
    else:
        sine = 1.0/sqrt(cosine*cosine - 1.0)

    u3x = sine * (u1y*u2z - u1z*u2y)
    u3y = sine * (u1z*u2x - u1x*u2z)
    u3z = sine * (u1x*u2y - u1y*u2x)

    u4x = cost * (u3y*u2z - u3z*u2y)
    u4y = cost * (u3z*u2x - u3x*u2z)
    u4z = cost * (u3x*u2y - u3y*u2x)

    return [x1 + dis*(-u2x*cosa + u4x + u3x*sint),
            y1 + dis*(-u2y*cosa + u4y + u3y*sint),
            z1 + dis*(-u2z*cosa + u4z + u3z*sint)]


def sc2vec(v1, v2, v3, dis=3.0):
    theta = 37.6 * DEGRAD

    # expand coords from passed vectors
    x1, y1, z1 = v1     # r(i-1), parent of next CA
    x2, y2, z2 = v2     # r(i), CA bonded to CB
    x3, y3, z3 = v3     # r(i+1), child of previous CA

#   print x1, y1, z1
#   print x2, y2, z2
#   print x3, y3, z3

    #          (ri - ri-1) + (ri - ri+1)
    #  X  =   ---------------------------
    #         |(ri - ri-1) + (ri - ri+1)|

    xX = x2 - x1
    xY = y2 - y1
    xZ = z2 - z1

    yX = x2 - x3
    yY = y2 - y3
    yZ = z2 - z3

    zX = xX + yX
    zY = xY + yY
    zZ = xZ + yZ
 

    zP = math.sqrt(zX*zX + zY*zY + zZ*zZ)
    zX/=zP
    zY/=zP
    zZ/=zP


    #          (ri - ri-1) * (ri - ri+1)
    #  Y  =   ---------------------------
    #         |(ri - ri-1) * (ri - ri+1)|


#   uX = xY*yZ - xZ*yY
#   uY = xZ*yX - xX*yZ
#   uZ = xX*yY - xY*yX
    uX = xX * yX
    uY = xY * yY
    uZ = xZ * yZ

    uP = math.sqrt(uX*uX + uY*uY + uZ*uZ)
    uX/=uP
    uY/=uP
    uZ/=uP


    vX = (zX * math.cos(theta) * dis) 
    vY = (zY * math.cos(theta) * dis)
    vZ = (zZ * math.cos(theta) * dis)

    wX = (uX * math.sin(theta) * dis)
    wY = (uY * math.sin(theta) * dis)
    wZ = (uZ * math.sin(theta) * dis)


    # R = dis * cos(theta * X) + dis * sin(theta * Y)

    CbX = (x2 + vX + wX) 
    CbY = (y2 + vY + wY) 
    CbZ = (z2 + vZ + wZ)

    return [CbX, CbY, CbZ]




def normalize_indices(rl):
    rp = map(len, rl)
    rp.insert(0, 0)
    rp = CUMSUM(rp)
    res = rl[0]
    for atom in res:
        atom.p1 = -1
    for i in range(1,len(rl)):
        res = rl[i]
        for atom in res:
            p = atom.p1
            if p < 0:
                atom.p1 = rp[i-1] + abs(p) - 1
            else:
                atom.p1 = rp[i] + p - 1

def read_seq_file(file):
    residues = []
    ff = open(file, 'r')
    nextline = ff.readline
    title = ''
    defalf = None
    deftau = None
    Resindx = 0
    while 1:
        line = nextline()
        if not line:
            break
        line = string.strip(line)
        if not line or line[:1] == '#': continue
        line = string.upper(line)
        if line[:3] == 'RES':
            sline = string.split(line)
            name = sline[1]
            args = sline[2:]
            res = {'name': name}


        # make a dictionary of residues and parameters which looks like,
        #{'ALF':  114.0, 'TAU':  -107.0, 'name': 'ALA'}
        #{'ALF':  114.0, 'TAU':  -107.0, 'name': 'ALA'}
        #{'ALF':  114.0, 'TAU':  -107.0, 'name': 'ALA'}
        #
        # Warning - not implemented yet - use DEF angles and dihedrals
        #
            if args:
                for i in range(0, len(args), 2):
                    res[args[i]] = string.atof(args[i+1])
            residues.append(res)
            Resindx = Resindx + 1
        elif line[:5] == 'TITLE':
            title = string.split(line)[1]
        elif line[:3] == 'DEF':
            sline = string.split(line)
            ang = sline[1]
            val = string.atof(sline[2])
            if ang == 'ALF':
                defalf = val
            elif ang == 'TAU':
                deftau = val
        else:
            raise ValueError,'Unknown keyword in input file: %s' % Split(line)[0]

    for res in residues:
        if res['name'] == 'END':
            res['ALF'] = 127.0
        elif not res.has_key('ALF'):
            if defalf:
                res['ALF'] = defalf
    for res in residues:
        if res['name'] == 'END':
            res['TAU'] = -163.0
        elif not res.has_key('TAU'):
                res['TAU'] = deftau
    return title, residues

def build_atom(line, Atom=Atom):
    args = string.split(string.upper(line))
    # name, p1, dist, alf, tau
    # name, p1, d1, d2, d3
    return Atom(args[0], string.atoi(args[4]), string.atof(args[1]),
                string.atof(args[2]), string.atof(args[3]))

def read_parm_file(file):
    ff = open(file, 'r')
    nextline = ff.readline
    Residues = {}
    while 1:
        line = nextline()
        if not line:
            break
        line = string.strip(line)
        line = string.upper(line)
        if line[:4] <> 'NAME':
            continue
        line = string.split(line)
        name = line[1]
        numatm = string.atoi(line[-1])
        newRes = Residue(name)
        for i in range(numatm):
            line = nextline()
            newRes.add_atom(build_atom(line))
        Residues[name] = newRes
    return Residues


def make_chain(rl, db):
    chain = []
    for res in rl:
        name = res['name']
        if not db.has_key(name):
            raise TypeError, 'Unknown Residue: %s' % name
        Res = db[name].copy()
        keys = res.keys()
        keys.remove('name')
        for atom in Res.atoms:
            if atom.name == 'CA':
                atom.d2 = res['ALF']
                atom.d3 = res['TAU']
#           print name,atom.name, atom.p1, atom.d1, atom.d2, atom.d3
        chain.append(Res)
    return chain

def get_CA_indx(atomlist):
    ca_atm_indices = []
    for i in range(len(atomlist)):
        if atomlist[i].name == 'CA':
            ca_atm_indices.append(i)

    return ca_atm_indices

def get_CB_indx(atomlist):
    cb_atm_indices = []
    for i in range(len(atomlist)):
        if atomlist[i].name == 'CB':
            cb_atm_indices.append(i)

    return cb_atm_indices

def build_pept(chain):
    normalize_indices(chain)
    # make backbone CA's first

    # first residue
    x1 = 1.0
    y1 = 1.0
    z1 = 1.0
    chain[0][0].xyz = ARRAY((x1, y1, z1), 'd')

    # second residue
    x2 = x1 + 3.80
    y2 = 1.0
    z2 = 1.0
    chain[1][0].xyz = ARRAY((x2, y2, z2), 'd')

    # third residue
    alf =  DEGRAD * chain[2][0].d2
    x3 = x2 + 3.80 * math.cos(math.pi - alf)
    y3 = y2 + 3.80 * math.sin(math.pi - alf)
    z3 = 1.0
    chain[2][0].xyz = ARRAY((x3, y3, z3), 'd')

    atomlist = []
    n = 0
    for res in chain:
        for atom in res.atoms:
            atomlist.append(atom)
#           print  n,atom.name,res.name,atom.p1, atom.d1, atom.d2, atom.d3
            n += 1

    ca_atm_indices = get_CA_indx(atomlist)

    # Build rest of CA's
    for i in xrange(3, len(ca_atm_indices)):
        a = atomlist[ca_atm_indices[i]]
        p1 = a.p1
        p2 = atomlist[p1].p1
        p3 = atomlist[p2].p1
        v1 = atomlist[p1].xyz
        v2 = atomlist[p2].xyz
        v3 = atomlist[p3].xyz
        dist = a.d1
        alf = atomlist[p1].d2
        tau = atomlist[p1].d3
        a.xyz = int2vec(v1, dist, v2, alf, v3, tau)
#       print ca_atm_indices[i], a.name,p1,p2,p3, alf, tau,a.xyz




    # Build CB's
    cb_atm_indices = get_CB_indx(atomlist)
    for i in xrange(len(cb_atm_indices)):
        a = atomlist[cb_atm_indices[i]]

        p1 =  a.p1
        p2 = atomlist[p1].p1
        p3 = p1 + 3
        v1 = atomlist[p1].xyz
        v2 = atomlist[p2].xyz
        v3 = atomlist[p3].xyz
        dis = a.d1
        alf = atomlist[p1].d2
        # most probable rotamer centroid is.
        ang = (360.0 - alf)/2.0 
        tor = a.d3
        a.xyz = int2vec(v1, dis, v2, ang, v3, tor)

    # Build backbone HB's
    for i in xrange(2, len(ca_atm_indices)):
        postCA = atomlist[ca_atm_indices[i]]
        preCA = atomlist[postCA.p1]
        HB = atomlist[ca_atm_indices[i]-1]
#       print ca_atm_indices[i],HB.name
        HBx = (postCA.xyz[0] + preCA.xyz[0]) / 2.0
        HBy = (postCA.xyz[1] + preCA.xyz[1]) / 2.0
        HBz = (postCA.xyz[2] + preCA.xyz[2]) / 2.0
        HB.xyz = [HBx, HBy, HBz]



def write_pdb(title, chain, file2):
    fmt = 'ATOM  %5d  %-4s%-4s %4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
    file2.write('COMPND    %s\n' % title)
    i = -1
    j = 0
    for res in chain:
        res_name = res.name
        for atom in res:
            i = i + 1
            name = atom.name
#           if name == 'CB': continue
#           if name == 'HB': continue
            file2.write(fmt % ( i, name, res_name, j, atom.xyz[0], atom.xyz[1],
                            atom.xyz[2], 1.0, 0.0))
        j = j + 1
    file2.write('END\n')
    file2.close()







def run_build(file1, file2):
    if file2 == None:
        file2 = sys.stdout
    else:
        file2 = open(file2, 'w')
    title, residues = read_seq_file(file1)
    parm_file = os.path.dirname(redux_2_1.__file__) + os.sep + 'Zmatrix.dat'
    db = read_parm_file(parm_file)
    chain = make_chain(residues, db)
    build_pept(chain)
    write_pdb(title, chain, file2)


if __name__ == "__main__":
    import sys, getopt
    args = sys.argv[1:]
    if len(args) == 1:
        filename1 = sys.argv[1]
        filename2 = None
    elif len(args) == 2:
        filename1 = sys.argv[1]
        filename2 = sys.argv[2]

    run_build(filename1, filename2)

