import string
import math
import os
import redux_2_1

DEGRAD = math.pi/180.0
RADDEG = 180.0/math.pi



class Atom:
    def __init__(self, name, p1, d1, d2, d3, d4):
        self.name = name
        self.p1 = p1        # first parent
        self.d1 = d1        # distance from first parent
        self.d2 = d2        # angle to p1 and p2
        self.d3 = d3        # torsion angle
        self.d4 = d4        # VDW radius

    def copy(self):
        return Atom(self.name, self.p1, self.d1, self.d2, self.d3, self.d4)


class Residue:
    def __init__(self, name, atoms=None):
        self.name = name
        if not atoms is None:
            self.atoms = atoms
        else:
            self.atoms = []
    def add_atom(self, atom):
        self.atoms.append(atom)

    def __getitem__(self, i):
        return self.atoms[i]

    def __len__(self):
        return len(self.atoms)

    def copy(self):
        newRes = Residue(self.name)
        for atom in self:
            newRes.add_atom(atom.copy())
        return newRes

def CUMSUM(a):
    l = [a[0]]
    for i in xrange(1, len(a)):
        l.append(a[i] + l[-1])
    return l

def angle(a1, a2, a3, acos=math.acos, sqrt=math.sqrt):
    """Angle in degrees between 3 atoms"""
    x1 = a1.x - a2.x
    y1 = a1.y - a2.y
    z1 = a1.z - a2.z

    x2 = a3.x - a2.x
    y2 = a3.y - a2.y
    z2 = a3.z - a2.z

    try:
        ang = (x1*x2 + y1*y2 + z1*z2)/sqrt((x1**2.0 + y1**2.0 + z1**2.0) *
                                           (x2**2.0 + y2**2.0 + z2**2.0))
    except ZeroDivisionError:
        return -999.999

    try:
        return acos(ang)*RADDEG
    except ValueError:
        if ang < -1.0:
            return 180.0
        else:
            return 0.0

def torsion(a1, a2, a3, a4, sqrt=math.sqrt, acos=math.acos):
    """Torsion angle in degrees between 4 atoms"""
    a = a1.x - a2.x; b = a1.y - a2.y; c = a1.z - a2.z;
    d = a3.x - a2.x; e = a3.y - a2.y; f = a3.z - a2.z;
    g = a4.x - a3.x; h = a4.y - a3.y; i = a4.z - a3.z;

    ax = b*f - e*c; ay = c*d - f*a; az = a*e - b*d;
    bx = h*f - e*i; by = i*d - f*g; bz = g*e - h*d;

    try:
        ang = (ax*bx+ay*by+az*bz)/sqrt((ax**2.0 + ay**2.0 + az**2.0) *
                                       (bx**2.0 + by**2.0 + bz**2.0))
    except ZeroDivisionError:
        return -999.999
    else:
        try:
            ang = acos(ang) * RADDEG
        except ValueError:
            if ang < -1.0:
                return 180.0
            else:
                return 0.0
        else:
            if ax*(e*bz-f*by) + ay*(f*bx-d*bz) + az*(d*by-e*bx) >= 0.0:
                return -ang
            else:
                return ang

def build_atom(line, Atom=Atom):
    args = string.split(string.upper(line))
    # name, p1, dist, alf, tau, 
    # name, p1, d1, d2, d3
    return Atom(args[0], string.atoi(args[4]), string.atof(args[1]),
                string.atof(args[2]), string.atof(args[3]), string.atof(args[5]))

def read_parm_file(file):
    ff = open(file, 'r')
    nextline = ff.readline
    Residues = {}
    while 1:
        line = nextline()
        if not line:
            break
        line = string.strip(line)
        line = string.upper(line)
        if line[:4] <> 'NAME':
            continue
        line = string.split(line)
        name = line[1]
        numatm = string.atoi(line[-1])
        newRes = Residue(name)
        for i in range(numatm):
            line = nextline()
            newRes.add_atom(build_atom(line))
        Residues[name] = newRes
    return Residues

def normalize_indices(rl):
    rp = map(len, rl)
    rp.insert(0, 0)
    rp = CUMSUM(rp)
    res = rl[0]
    for atom in res:
        atom.p1 = -1
    for i in range(1,len(rl)):
        res = rl[i]
        for atom in res:
            p = atom.p1
            if p < 0:
                atom.p1 = rp[i-1] + abs(p) - 1
            else:
                atom.p1 = rp[i] + p - 1

def protein_from_pdb(file):
    parm_file = os.path.dirname(redux_2_1.__file__) + os.sep + 'Zmatrix.dat'
    db = read_parm_file(parm_file)
    data = open(file, 'rb').readlines()
    atoms = []
    res_names = []
    res_first_atom_indices = []
    res_pdb_number = []
    prot_name = ' '
    na = nr = -1
    current_res = None

    # get res names and make chain
    rl = []
    atom_rec = []
    for line in data:
        if not line:
            break
        if line[:4] == 'ATOM':
            atom_rec.append((line))
            rnam = line[17:20]
            rnum = line[22:27]
            if rnum <> current_res:
                rl.append(rnam)
                current_res = rnum
        elif line[:4] == 'COMP':
            prot_name = string.strip(line[10:70])
        elif line[:3] in ('TER', 'END'):
            break


    chain = []
    for i in range(len(rl)):
        name = rl[i]
        if not db.has_key(name):
            raise TypeError, 'Unknown Residue: %s' % name
        Res = db[name].copy()
        chain.append(Res)

    normalize_indices(chain)

    # make atomlist from chain
    atomlist = []
    for res in chain:
        for atom in res.atoms:
            atomlist.append(atom)

    # Remove last HB
    # This can't be made from real PDB files and so 
    # just ignore it for energies
    atomlist.pop()

    for i in range(len(atomlist)):
        atomlist[i].indx = i
        atomlist[i].rnam = atom_rec[i][17:20]
        atomlist[i].rnum = atom_rec[i][22:27]
        atomlist[i].x = float(atom_rec[i][30:38])
        atomlist[i].y = float(atom_rec[i][38:46])
        atomlist[i].z = float(atom_rec[i][46:54])
#       print atomlist[i].indx,atomlist[i].name, atom_rec[i][22:27]


    ca_atoms = []
    for i in range(len(atomlist)):
        if atomlist[i].name == 'CA':
            ca_atoms.append(atomlist[i])


    for i in range(1, len(ca_atoms)-1):
        ang = angle(ca_atoms[i-1], ca_atoms[i], ca_atoms[i+1])
        ca_atoms[i].d2 = ang

    for i in range(2, len(ca_atoms)-1):
        tor = torsion(ca_atoms[i-2],ca_atoms[i-1],ca_atoms[i],ca_atoms[i+1])
        ca_atoms[i].d3 = tor


#   for i in range(len(atomlist)):
#       a = atomlist[i]
#       print a.indx,a.name,a.rnam,a.rnum,a.x,a.y,a.z,a.p1,a.d1,a.d2,a.d3,a.d4

    return prot_name,atomlist

def write_pdb(prot_name, atomlist, filename):
    if type(filename) is type(''):
        ff = open(filename, 'wb')
    else:
        ff = filename

    write = ff.write

    write('COMPND     %s\n' % prot_name)

    pdbfmt = 'ATOM  %5d  %-3s %-3s  %4d    %8.3f%8.3f%8.3f  0.00  0.00\n'
    for i in xrange(len(atomlist)):
        a = atomlist[i]
        write(pdbfmt % (a.indx,a.name,a.rnam,int(a.rnum),a.x,a.y,a.z))
    write('TER\n')

    ff.close()


def MakeMol(file):
    prot_name, atomlist = protein_from_pdb(file)
    for i in range(len(atomlist)):
        a = atomlist[i]
        print a.indx,a.name,a.rnam,a.rnum,a.x,a.y,a.z,a.p1,a.d1,a.d2,a.d3,a.d4


#prot_name, atomlist = protein_from_pdb('all.pdb')
#write_pdb(prot_name, atomlist, 'new.pdb')


if __name__ == "__main__":
    import sys
    filename = sys.argv[1]
    MakeMol(filename)

