import string
import math
import os
import redux_2_1

DEGRAD = math.pi/180.0
RADDEG = 180.0/math.pi



class Atom:
    def __init__(self, name, p1, d1, d2, d3, d4):
        self.name = name
        self.p1 = p1        # first parent
        self.d1 = d1        # distance from first parent
        self.d2 = d2        # angle to p1 and p2
        self.d3 = d3        # torsion angle
        self.d4 = d4        # VDW radius

    def copy(self):
        return Atom(self.name, self.p1, self.d1, self.d2, self.d3, self.d4)


class Residue:
    def __init__(self, name, atoms=None):
        self.name = name
        if not atoms is None:
            self.atoms = atoms
        else:
            self.atoms = []
    def add_atom(self, atom):
        self.atoms.append(atom)

    def __getitem__(self, i):
        return self.atoms[i]

    def __len__(self):
        return len(self.atoms)

    def copy(self):
        newRes = Residue(self.name)
        for atom in self:
            newRes.add_atom(atom.copy())
        return newRes

def CUMSUM(a):
    l = [a[0]]
    for i in xrange(1, len(a)):
        l.append(a[i] + l[-1])
    return l

def angle(a1, a2, a3, acos=math.acos, sqrt=math.sqrt):
    """Angle in degrees between 3 atoms"""
    x1 = a1.x - a2.x
    y1 = a1.y - a2.y
    z1 = a1.z - a2.z

    x2 = a3.x - a2.x
    y2 = a3.y - a2.y
    z2 = a3.z - a2.z

    try:
        ang = (x1*x2 + y1*y2 + z1*z2)/sqrt((x1**2.0 + y1**2.0 + z1**2.0) *
                                           (x2**2.0 + y2**2.0 + z2**2.0))
    except ZeroDivisionError:
        return -999.999

    try:
        return acos(ang)*RADDEG
    except ValueError:
        if ang < -1.0:
            return 180.0
        else:
            return 0.0

def torsion(a1, a2, a3, a4, sqrt=math.sqrt, acos=math.acos):
    """Torsion angle in degrees between 4 atoms"""
    a = a1.x - a2.x; b = a1.y - a2.y; c = a1.z - a2.z;
    d = a3.x - a2.x; e = a3.y - a2.y; f = a3.z - a2.z;
    g = a4.x - a3.x; h = a4.y - a3.y; i = a4.z - a3.z;

    ax = b*f - e*c; ay = c*d - f*a; az = a*e - b*d;
    bx = h*f - e*i; by = i*d - f*g; bz = g*e - h*d;

    try:
        ang = (ax*bx+ay*by+az*bz)/sqrt((ax**2.0 + ay**2.0 + az**2.0) *
                                       (bx**2.0 + by**2.0 + bz**2.0))
    except ZeroDivisionError:
        return -999.999
    else:
        try:
            ang = acos(ang) * RADDEG
        except ValueError:
            if ang < -1.0:
                return 180.0
            else:
                return 0.0
        else:
            if ax*(e*bz-f*by) + ay*(f*bx-d*bz) + az*(d*by-e*bx) >= 0.0:
                return -ang
            else:
                return ang

def int2vec(v1, dis, v2, ang, v3, tors, sin=math.sin,\
            cos=math.cos, sqrt=math.sqrt):
    """int2vec(v1, dis, v2, ang, v3, tors)
        Compute coordinates for vector v, given its 'dis'tance
        from vector v1, 'ang'le v-v1-v2 and 'tors'ion v-v1-v2-v3.'
        Both ang and tors should be in degrees.
    """

    ang = ang*DEGRAD
    sina = sin(ang)
    cosa = cos(ang)

    tors = (tors*DEGRAD)
    sint = sina * sin(tors)
    cost = sina * cos(tors)

    x1, y1, z1 = v1
    x2, y2, z2 = v2
    x3, y3, z3 = v3

    u1x = x2 - x3
    u1y = y2 - y3
    u1z = z2 - z3
    d = 1.0 / sqrt(u1x*u1x + u1y*u1y + u1z*u1z)
    u1x = u1x*d
    u1y = u1y*d
    u1z = u1z*d

    u2x = x1 - x2
    u2y = y1 - y2
    u2z = z1 - z2;
    d = 1.0 / sqrt(u2x*u2x + u2y*u2y + u2z*u2z)
    u2x = u2x*d
    u2y = u2y*d
    u2z = u2z*d

    cosine = u1x*u2x + u1y*u2y + u1z*u2z

    if (abs(cosine) < 1.0):
        sine = 1.0/sqrt(1.0 - cosine*cosine)
    else:
        sine = 1.0/sqrt(cosine*cosine - 1.0)

    u3x = sine * (u1y*u2z - u1z*u2y)
    u3y = sine * (u1z*u2x - u1x*u2z)
    u3z = sine * (u1x*u2y - u1y*u2x)

    u4x = cost * (u3y*u2z - u3z*u2y)
    u4y = cost * (u3z*u2x - u3x*u2z)
    u4z = cost * (u3x*u2y - u3y*u2x)

    return [x1 + dis*(-u2x*cosa + u4x + u3x*sint),
            y1 + dis*(-u2y*cosa + u4y + u3y*sint),
            z1 + dis*(-u2z*cosa + u4z + u3z*sint)]


def build_atom(line, Atom=Atom):
    args = string.split(string.upper(line))
    # name, p1, dist, alf, tau, 
    # name, p1, d1, d2, d3
    return Atom(args[0], string.atoi(args[4]), string.atof(args[1]),
                string.atof(args[2]), string.atof(args[3]), string.atof(args[5]))

def read_parm_file(file):
    ff = open(file, 'r')
    nextline = ff.readline
    Residues = {}
    while 1:
        line = nextline()
        if not line:
            break
        line = string.strip(line)
        line = string.upper(line)
        if line[:4] <> 'NAME':
            continue
        line = string.split(line)
        name = line[1]
        numatm = string.atoi(line[-1])
        newRes = Residue(name)
        for i in range(numatm):
            line = nextline()
            newRes.add_atom(build_atom(line))
        Residues[name] = newRes
    return Residues

def normalize_indices(rl):
    rp = map(len, rl)
    rp.insert(0, 0)
    rp = CUMSUM(rp)
    res = rl[0]
    for atom in res:
        atom.p1 = -1
    for i in range(1,len(rl)):
        res = rl[i]
        for atom in res:
            p = atom.p1
            if p < 0:
                atom.p1 = rp[i-1] + abs(p) - 1
            else:
                atom.p1 = rp[i] + p - 1

def protein_from_pdb(file):
    parm_file = os.path.dirname(redux_2_1.__file__) + os.sep + 'Zmatrix.dat'
    db = read_parm_file(parm_file)
    data = open(file, 'rb').readlines()
    atoms = []
    res_names = []
    res_first_atom_indices = []
    res_pdb_number = []
    prot_name = ' '
    na = nr = -1
    current_res = None

    # get res names and make chain
    rl = []
    atom_rec = []
    for line in data:
        if line[:4] == 'ATOM' and (line[13:15] == 'CA' or line[13:15] == 'CB'):
            atom_rec.append((line))
            rnam = line[17:20]
            rnum = line[22:27]
            if rnum <> current_res:
                rl.append(rnam)
                current_res = rnum
            # Get dummy place holder for HB atoms
            if (line[17:20] == 'GLY' and line[13:15] == 'CA') or line[13:15] == 'CB':
                atom_rec.append((line))
        elif line[:4] == 'COMP':
            prot_name = string.strip(line[10:70])
        elif line[:3] in ('TER', 'END'):
            break

#   for i in range(len(atom_rec)):
#       print atom_rec[i]

    chain = []
    for i in range(len(rl)):
        name = rl[i]
        if not db.has_key(name):
            raise TypeError, 'Unknown Residue: %s' % name
        Res = db[name].copy()
        chain.append(Res)

    normalize_indices(chain)

    # make atomlist from chain
    atomlist = []
    for res in chain:
        for atom in res.atoms:
            atomlist.append(atom)
  
    for i in range(len(atomlist)):
#       print atomlist[i].name, atom_rec[i]
        try:

            atomlist[i].indx = i
            atomlist[i].rnam = atom_rec[i][17:20]
            atomlist[i].rnum = atom_rec[i][22:27]
            atomlist[i].x = float(atom_rec[i][30:38])
            atomlist[i].y = float(atom_rec[i][38:46])
            atomlist[i].z = float(atom_rec[i][46:54])
        except:
            print ' Problem: ', file

    # Remove last HB
    atomlist.pop()

    # Remove first HB
#   if atomlist[1].name == 'HB':
#       atomlist.pop(1)
#   elif atomlist[2].name == 'HB':
#       atomlist.pop(2)



#   for i in range(len(atomlist)):
#       print atomlist[i].name, atomlist[i].rnam,atomlist[i].rnum

    ca_atoms = []
    for i in range(len(atomlist)):
        if atomlist[i].name == 'CA':
            ca_atoms.append(atomlist[i])


    for i in range(1, len(ca_atoms)-1):
        ang = angle(ca_atoms[i-1], ca_atoms[i], ca_atoms[i+1])
        ca_atoms[i].d2 = ang

    for i in range(2, len(ca_atoms)-1):
        tor = torsion(ca_atoms[i-2],ca_atoms[i-1],ca_atoms[i],ca_atoms[i+1])
        ca_atoms[i].d3 = tor

    for i in range(len(atomlist)):
        for j in range(len(ca_atoms)):
            if ca_atoms[j].indx == atomlist[i].indx:
                atomlist[i].d2 == ca_atoms[j].d2
                atomlist[i].d3 == ca_atoms[j].d3

    cb_atoms = []
    for i in range(len(atomlist)):
        if atomlist[i].name == 'CB':
            cb_atoms.append(atomlist[i])

    for i in xrange(3, len(atomlist)-1):
        a = atomlist[i]
        if a.name == 'CB':
            p1 =  a.p1
            p2 = atomlist[p1].p1
            p3 = p1 + 3
            v1 = []
            v1.append(atomlist[p1].x)
            v1.append(atomlist[p1].y)
            v1.append(atomlist[p1].z)
            v2 = []
            v2.append(atomlist[p2].x)
            v2.append(atomlist[p2].y)
            v2.append(atomlist[p2].z)
            v3 = []
            v3.append(atomlist[p3].x)
            v3.append(atomlist[p3].y)
            v3.append(atomlist[p3].z)
            dis = a.d1
            alf = atomlist[p1].d2
            ang = (360.0 - alf)/2.0
            tor = a.d3
            xyz = int2vec(v1, dis, v2, ang, v3, tor)
            a.x = xyz[0]
            a.y = xyz[1]
            a.z = xyz[2]
        elif a.name == 'CA':
            continue

    ca_atm_indices = []
    for i in range(len(atomlist)):
        if atomlist[i].name == 'CA':
            ca_atm_indices.append(atomlist[i].indx)

    # Build backbone HB's
    for i in xrange(1, len(ca_atm_indices)):
        postCA = atomlist[ca_atm_indices[i]]
        preCA = atomlist[postCA.p1]
        HB = atomlist[ca_atm_indices[i]-1]
        HB.x = (postCA.x + preCA.x) / 2.0
        HB.y = (postCA.y + preCA.y) / 2.0
        HB.z = (postCA.z + preCA.z) / 2.0


#   for i in range(len(atomlist)):
#       a = atomlist[i]
#       print a.indx,a.name,a.rnam,a.rnum,a.x,a.y,a.z,a.p1,a.d1,a.d2,a.d3,a.d4

    return prot_name,atomlist

def write_pdb(prot_name, atomlist, filename):
    if type(filename) is type(''):
        ff = open(filename, 'wb')
    else:
        ff = filename

    write = ff.write

    write('COMPND     %s\n' % prot_name)

    pdbfmt = 'ATOM  %5d  %-3s %-3s  %4d    %8.3f%8.3f%8.3f  0.00  0.00\n'
    for i in xrange(len(atomlist)):
        a = atomlist[i]
        write(pdbfmt % (a.indx,a.name,a.rnam,int(a.rnum),a.x,a.y,a.z))
    write('TER\n')

    ff.close()

def distance(a, b):
    """Euclidean distance between two atoms"""
    return math.sqrt((a.x - b.x)**2.0 + (a.y - b.y)**2.0 + (a.z - b.z)**2.0)

def get_cb_dist(atomlist,prot_name):
    cb_atoms = []
    for i in range(len(atomlist)):
        if atomlist[i].rnam == 'GLY' and atomlist[i].name == 'CA':
            cb_atoms.append(atomlist[i])
        if atomlist[i].name == 'CB':
            cb_atoms.append(atomlist[i])

    for i in range(len(cb_atoms)-3):
        for j in range(i+3, len(cb_atoms)):
            dist = distance(cb_atoms[i], cb_atoms[j])
            if dist < 8.0:
                pass
                print cb_atoms[i].rnam,cb_atoms[i].rnum, \
                      cb_atoms[j].rnam,cb_atoms[j].rnum, dist, prot_name




def MakeMol(file):
    prot_name, atomlist = protein_from_pdb(file)
    basename = os.path.splitext(file)[0][-4:]
    write_pdb(prot_name, atomlist, basename + '_redux.pdb')



if __name__ == "__main__":
    import sys
    filename = sys.argv[1]
    MakeMol(filename)

