"""
Makes a Redux structure from a file listing sequence and phi, psi angles.

Format of the input seq_ang file is,

TITLE      Test
RES ALA PHI -60.0 PSI -40.0
RES LEU PHI -120.0 PSI 120.0
RES ILE PHI -120.0 PSI 120.0

This requires a Zmatrix

Default is to write to the screen if no output file name is given.
"""
USAGE = """
python Redux_from_Seq_Ang.py -i test.seq_ang  -p Zmatrix.dat [-o test.pdb]
"""
import string
import math
import os

try:
    import Numeric
except ImportError:
    def CUMSUM(a):
        l = [a[0]]
        for i in xrange(1, len(a)):
            l.append(a[i] + l[-1])
        return l
    def ZEROS(n, dummy):
        return [0.0]*n
    def ARRAY(v, dummy):
        return list(v)
else:
    CUMSUM = Numeric.add.accumulate
    ZEROS = Numeric.zeros
    ARRAY = Numeric.array

DEGRAD = math.pi/180.0
RADDEG = 180.0/math.pi

class Atom:
    def __init__(self, name, p1, d1, d2, d3, d4):
        self.name = name
        self.p1 = p1        # first parent
        self.d1 = d1        # distance from first parent
        self.d2 = d2        # angle to p1 and p2
        self.d3 = d3        # torsion angle
        self.d4 = d4        # VDW radius

    def copy(self):
        return Atom(self.name, self.p1, self.d1, self.d2, self.d3, self.d4)


class Residue:
    def __init__(self, name, atoms=None):
        self.name = name
        if not atoms is None:
            self.atoms = atoms
        else:
            self.atoms = []
    def add_atom(self, atom):
        self.atoms.append(atom)

    def __getitem__(self, i):
        return self.atoms[i]

    def __len__(self):
        return len(self.atoms)

    def copy(self):
        newRes = Residue(self.name)
        for atom in self:
            newRes.add_atom(atom.copy())
        return newRes

def angle(a1, a2, a3, acos=math.acos, sqrt=math.sqrt):
    """Angle in degrees between 3 atoms"""
    x1 = a1.x - a2.x
    y1 = a1.y - a2.y
    z1 = a1.z - a2.z

    x2 = a3.x - a2.x
    y2 = a3.y - a2.y
    z2 = a3.z - a2.z

    try:
        ang = (x1*x2 + y1*y2 + z1*z2)/sqrt((x1**2.0 + y1**2.0 + z1**2.0) *
                                           (x2**2.0 + y2**2.0 + z2**2.0))
    except ZeroDivisionError:
        return -999.999

    try:
        return acos(ang)*RADDEG
    except ValueError:
        if ang < -1.0:
            return 180.0
        else:
            return 0.0

def torsion(a1, a2, a3, a4, sqrt=math.sqrt, acos=math.acos):
    """Torsion angle in degrees between 4 atoms"""
    a = a1.x - a2.x; b = a1.y - a2.y; c = a1.z - a2.z;
    d = a3.x - a2.x; e = a3.y - a2.y; f = a3.z - a2.z;
    g = a4.x - a3.x; h = a4.y - a3.y; i = a4.z - a3.z;

    ax = b*f - e*c; ay = c*d - f*a; az = a*e - b*d;
    bx = h*f - e*i; by = i*d - f*g; bz = g*e - h*d;

    try:
        ang = (ax*bx+ay*by+az*bz)/sqrt((ax**2.0 + ay**2.0 + az**2.0) *
                                       (bx**2.0 + by**2.0 + bz**2.0))
    except ZeroDivisionError:
        return -999.999
    else:
        try:
            ang = acos(ang) * RADDEG
        except ValueError:
            if ang < -1.0:
                return 180.0
            else:
                return 0.0
        else:
            if ax*(e*bz-f*by) + ay*(f*bx-d*bz) + az*(d*by-e*bx) >= 0.0:
                return -ang
            else:
                return ang



class FullAtom:
    def __init__(self, name, p1, p2, p3, d1, d2, d3, iscore, chitype):
        self.name = name
        self.p1 = p1        # first parent
        self.p2 = p2        # second parent
        self.p3 = p3        # third parent
        self.d1 = d1        # distance from first parent
        self.d2 = d2        # angle made with first and second parents
        self.d3 = d3        # torsion made withe first, second and third parents
        self.iscore = (iscore == '+' or iscore == 1)
        self.chitype = chitype

    def copy(self):
        return FullAtom(self.name, self.p1, self.p2, self.p3, self.d1,
                    self.d2, self.d3, self.iscore, self.chitype)


class FullResidue:
    def __init__(self, name, atoms=None):
        self.name = name
        if not atoms is None:
            self.atoms = atoms
        else:
            self.atoms = []
    def add_atom(self, atom):
        self.atoms.append(atom)

    def __getitem__(self, i):
        return self.atoms[i]

    def __len__(self):
        return len(self.atoms)

    def chiatom(self, chiname):
        for atom in self.atoms:
            if atom.chitype == chiname:
                return atom

    def copy(self):
        newRes = FullResidue(self.name)
        for atom in self:
            newRes.add_atom(atom.copy())
        return newRes

def int2vec(v1, dis, v2, ang, v3, tors, sin=math.sin,\
            cos=math.cos, sqrt=math.sqrt):
    """int2vec(v1, dis, v2, ang, v3, tors)
        Compute coordinates for vector v, given its 'dis'tance
        from vector v1, 'ang'le v-v1-v2 and 'tors'ion v-v1-v2-v3.'
        Both ang and tors should be in degrees.
    """

    DEG2RAD = math.pi/180
    ang = ang*DEG2RAD
    sina = sin(ang)
    cosa = cos(ang)

    tors = (tors*DEG2RAD)
    sint = sina * sin(tors)
    cost = sina * cos(tors)

    x1, y1, z1 = v1
    x2, y2, z2 = v2
    x3, y3, z3 = v3

    u1x = x2 - x3
    u1y = y2 - y3
    u1z = z2 - z3
    d = 1.0 / sqrt(u1x*u1x + u1y*u1y + u1z*u1z)
    u1x = u1x*d
    u1y = u1y*d
    u1z = u1z*d

    u2x = x1 - x2
    u2y = y1 - y2
    u2z = z1 - z2;
    d = 1.0 / sqrt(u2x*u2x + u2y*u2y + u2z*u2z)
    u2x = u2x*d
    u2y = u2y*d
    u2z = u2z*d

    cosine = u1x*u2x + u1y*u2y + u1z*u2z

    if (abs(cosine) < 1.0):
        sine = 1.0/sqrt(1.0 - cosine*cosine)
    else:
        sine = 1.0/sqrt(cosine*cosine - 1.0)

    u3x = sine * (u1y*u2z - u1z*u2y)
    u3y = sine * (u1z*u2x - u1x*u2z)
    u3z = sine * (u1x*u2y - u1y*u2x)

    u4x = cost * (u3y*u2z - u3z*u2y)
    u4y = cost * (u3z*u2x - u3x*u2z)
    u4z = cost * (u3x*u2y - u3y*u2x)

    return [x1 + dis*(-u2x*cosa + u4x + u3x*sint),
            y1 + dis*(-u2y*cosa + u4y + u3y*sint),
            z1 + dis*(-u2z*cosa + u4z + u3z*sint)]


def build_atom(line, Atom=Atom):
    args = string.split(string.upper(line))
    # name, p1, dist, alf, tau,
    # name, p1, d1, d2, d3
    return Atom(args[0], string.atoi(args[4]), string.atof(args[1]),
                string.atof(args[2]), string.atof(args[3]), string.atof(args[5]))


def read_parm_file(file):
    ff = open(file, 'r')
    nextline = ff.readline
    Residues = {}
    while 1:
        line = nextline()
        if not line:
            break
        line = string.strip(line)
        line = string.upper(line)
        if line[:4] <> 'NAME':
            continue
        line = string.split(line)
        name = line[1]
        numatm = string.atoi(line[-1])
        newRes = Residue(name)
        for i in range(numatm):
            line = nextline()
            newRes.add_atom(build_atom(line))
        Residues[name] = newRes
    return Residues

def normalize_indices(rl):
    rp = map(len, rl)
    rp.insert(0, 0)
    rp = CUMSUM(rp)
    res = rl[0]
    for atom in res:
        atom.p1 = -1
    for i in range(1,len(rl)):
        res = rl[i]
        for atom in res:
            p = atom.p1
            if p < 0:
                atom.p1 = rp[i-1] + abs(p) - 1
            else:
                atom.p1 = rp[i] + p - 1


#
def Full_build_atom(line, split=string.split, atof=string.atof, atoi=string.atoi,
                       upper=string.upper, FullAtom=FullAtom):
    args = split(upper(line))
    return FullAtom(args[0], atoi(args[4]), atoi(args[5]), atoi(args[6]),
                atof(args[1]), atof(args[2]), atof(args[3]), args[7],
                args[8])


def _read_input(file):
    # Makes a list of dictionaries. One dict for each residue. 
    # For example ALA ALA, where the phi and psi values have been specified.
    # residues = [{'PHI': -60.0, 'PSI': 140.0, 'name': 'ALA'}, {'PHI': -120.0, 'PSI': -60.0, 'name': 'ALA'}]
    residues = []
    if type(file) == type(''):
        ff = open(file, 'r')
    else:
        ff = file
    nextline = ff.readline
    allh = 0
    title = ''
    defphi = None
    defpsi = None
    Strip = string.strip
    Split = string.split
    Upper = string.upper
    Float = string.atof
    Resindx = 0
    while 1:
        line = nextline()
        if not line:
            break
        line = Strip(line)
        if not line or line[:1] == '#': continue
        line = Upper(line)
        if line[:3] == 'RES':
            sline = Split(line)
            name = sline[1]
            args = sline[2:]
#           print args
            res = {'name': name}
#           print res
            if args:
                for i in range(0, len(args), 2):
                    Ome = 180.0
                    res[args[i]] = Float(args[i+1])
            residues.append(res)
            Resindx = Resindx + 1
        elif line[:5] == 'TITLE':
            title = Split(line)[1]
        else:
            raise ValueError,'Unknown keyword in input file: %s' % Split(line)[0]

    for res in residues:
        val = res['PSI'] + 180.0
        if val > 180:
            res['PSI'] = val - 360.
        else:
            res['PSI'] = val
#       print res
#   print residues
    return title, residues



def Full_read_parm_file(file):
    if type(file) == type(''):
        ff = open(file, 'r')
    else:
        ff = file
    nextline = ff.readline
    Strip = string.strip
    Split = string.split
    Upper = string.upper
    Int = string.atoi
        
    FullResidues = {}
    while 1:  
        line = nextline()
        if not line:
            break
        line = Strip(line)
        line = Upper(line)
#       if line[:4] <> 'NAME':
        if line[5:8] <> 'DUM':
            continue
        line = Split(line)
        name = line[1]
        numatm = Int(line[-1])
        newRes = FullResidue(name)
        for i in range(numatm):
            line = nextline()
            newRes.add_atom(Full_build_atom(line))
        FullResidues[name] = newRes
    return FullResidues

def make_chain(rl, db):
    # rl = residue list
    # db = FullResidues
    chain = []
    for res in rl:
        name = res['name']
#       if not db.has_key(name):
#           raise TypeError, 'Unknown FullResidue: %s' % name
        Res = db['DUM'].copy()
        Res.name = name
        keys = res.keys()
        keys.remove('name')
        for key in keys:
            val = res[key]
            for atom in Res.atoms:
                if atom.chitype == key:
                    atom.d3 = val
        chain.append(Res)
    return chain

def Full_normalize_indices(rl):
    rp = map(len, rl)
    rp.insert(0, 0)
    rp = CUMSUM(rp)
    for i in range(len(rl)):
        res = rl[i]
        for atom in res:
            p = atom.p1
            if p < 0:
                atom.p1 = rp[i-1] + abs(p) - 1
            else:
                atom.p1 = rp[i] + p - 1
            p = atom.p2
            if p < 0:
                atom.p2 = rp[i-1] + abs(p) - 1
            else:
                atom.p2 = rp[i] + p - 1
            p = atom.p3
            if p < 0:
                atom.p3 = rp[i-1] + abs(p) - 1
            else:
                atom.p3 = rp[i] + p - 1
def build_peptide(rl):
    Full_normalize_indices(rl)
    dtor = math.pi/180.0
    res0 = rl[0]
    res0[0].xyz = ZEROS(3, 'd')
    a2 = res0[1]
    res0[1].xyz = ARRAY((a2.d1, 0, 0), 'd')
    a3 = res0[2]
    if a3.p1 == 1:
        x = a2.d1 - a3.d1*math.cos(dtor*a3.d2)
    else:
        x = a3.d1*math.cos(dtor*a3.d2)
    y = a3.d1 * math.sin(a3.d2*dtor)
    res0[2].xyz = ARRAY((x, y, 0.0), 'd')
    atomlist = []
    for res in rl:
        for atom in res.atoms:
            atomlist.append(atom)

    for i in xrange(3, len(atomlist)):
        a = atomlist[i]
        p1 = atomlist[a.p1].xyz
        p2 = atomlist[a.p2].xyz
        p3 = atomlist[a.p3].xyz
        a.xyz = int2vec(p1, a.d1, p2, a.d2, p3, a.d3)

def read_parm_file(file):
    ff = open(file, 'r')
    nextline = ff.readline
    Residues = {}
    while 1:
        line = nextline()
        if not line:
            break
        line = string.strip(line)
        line = string.upper(line)
        if line[:4] <> 'NAME':
            continue
        line = string.split(line)
        name = line[1]
        numatm = string.atoi(line[-1])
        newRes = Residue(name)
        for i in range(numatm):
            line = nextline()
            newRes.add_atom(build_atom(line))
        Residues[name] = newRes
    return Residues

def redux_from_full(parm_file, full_chain):
#   parm_file = 'Zmatrix.dat'
    DB = read_parm_file(parm_file)
    ATOMS = []
    RES_NAMES = []
    RES_FIRST_ATOM_INDICES = []
    RES_PDB_NUMBER = []
    PROT_NAME = ' '
    NA = NR = -1
    CURRENT_RES = None
    RL = []
    ATOM_REC = []

    fmt1 = 'ATOM  %5d  %-4s%-4s %4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
    i = 1
    j = 1
    for res in full_chain:
        res_name = res.name
        for atom in res:
            name = atom.name
            if name == 'DU': continue

#           sys.stdout.write(fmt1 % ( i, name, res_name, j, atom.xyz[0], atom.xyz[1],
#                           atom.xyz[2], 1.0, 0.0))
            if name == 'CA' or name == 'CB':
                RNAM = res_name
                RNUM = j
                if RNUM <> CURRENT_RES:
                    RL.append(RNAM)
                    CURRENT_RES = RNUM
                ATOM_REC.append([name,res_name, j, atom.xyz[0], atom.xyz[1],atom.xyz[2] ])
            # Get dummy place holder for HB atoms
            if (res_name == 'GLY' and name == 'CA') or name == 'CB':
                ATOM_REC.append([name,res_name, j, atom.xyz[0], atom.xyz[1],atom.xyz[2] ])
            i = i + 1
        j = j + 1

#   for i in range(len(ATOM_REC)):
#       print ATOM_REC[i]

    CHAIN = []
    for i in range(len(RL)):
        NAME = RL[i]
        RES = DB[NAME].copy()
        CHAIN.append(RES)
    
    normalize_indices(CHAIN)

    # make atomlist from chain
    ATOMLIST = []
    for RES in CHAIN:
        for ATOM in RES.atoms:
            ATOMLIST.append(ATOM)

    for i in range(len(ATOMLIST)):
#       print ATOMLIST[i].name
#       print len(ATOMLIST), len(ATOM_REC)
        try:

            ATOMLIST[i].indx = i
            ATOMLIST[i].rnam = ATOM_REC[i][1]
            ATOMLIST[i].rnum = ATOM_REC[i][2]
            ATOMLIST[i].x = ATOM_REC[i][3]
            ATOMLIST[i].y = ATOM_REC[i][4]
            ATOMLIST[i].z = ATOM_REC[i][5]
        except:
            print ' Problem: '

    # Remove last HB
    ATOMLIST.pop()

#   for i in range(len(ATOMLIST)):
#       print ATOMLIST[i].name,ATOMLIST[i].rnam,ATOMLIST[i].rnum,ATOMLIST[i].x,ATOMLIST[i].y,ATOMLIST[i].z

    ca_atoms = []
    for i in range(len(ATOMLIST)):
        if ATOMLIST[i].name == 'CA':
            ca_atoms.append(ATOMLIST[i])

    for i in range(1, len(ca_atoms)-1):
        ang = angle(ca_atoms[i-1], ca_atoms[i], ca_atoms[i+1])
        ca_atoms[i].d2 = ang

    for i in range(2, len(ca_atoms)-1):
        tor = torsion(ca_atoms[i-2],ca_atoms[i-1],ca_atoms[i],ca_atoms[i+1])
        ca_atoms[i].d3 = tor

    for i in range(len(ATOMLIST)):
        for j in range(len(ca_atoms)):
            if ca_atoms[j].indx == ATOMLIST[i].indx:
                ATOMLIST[i].d2 == ca_atoms[j].d2
                ATOMLIST[i].d3 == ca_atoms[j].d3

    cb_atoms = []
    for i in range(len(ATOMLIST)):
        if ATOMLIST[i].name == 'CB':
            cb_atoms.append(ATOMLIST[i])

    for i in xrange(3, len(ATOMLIST)-1):
        a = ATOMLIST[i]
        if a.name == 'CB':
            p1 =  a.p1
            p2 = ATOMLIST[p1].p1
            p3 = p1 + 3
            v1 = []
            v1.append(ATOMLIST[p1].x)
            v1.append(ATOMLIST[p1].y)
            v1.append(ATOMLIST[p1].z)
            v2 = []
            v2.append(ATOMLIST[p2].x)
            v2.append(ATOMLIST[p2].y)
            v2.append(ATOMLIST[p2].z)
            v3 = []
            v3.append(ATOMLIST[p3].x)
            v3.append(ATOMLIST[p3].y)
            v3.append(ATOMLIST[p3].z)
            dis = a.d1
            alf = ATOMLIST[p1].d2
            ang = (360.0 - alf)/2.0
            tor = a.d3
            xyz = int2vec(v1, dis, v2, ang, v3, tor)
            a.x = xyz[0]
            a.y = xyz[1]
            a.z = xyz[2]
        elif a.name == 'CA':
            continue

    ca_atm_indices = []
    for i in range(len(ATOMLIST)):
        if ATOMLIST[i].name == 'CA':
            ca_atm_indices.append(ATOMLIST[i].indx)

    # Build backbone HB's
    for i in xrange(1, len(ca_atm_indices)):
        postCA = ATOMLIST[ca_atm_indices[i]]
        preCA = ATOMLIST[postCA.p1]
        HB = ATOMLIST[ca_atm_indices[i]-1]
        HB.x = (postCA.x + preCA.x) / 2.0
        HB.y = (postCA.y + preCA.y) / 2.0
        HB.z = (postCA.z + preCA.z) / 2.0

    return ATOMLIST

def write_pdb(title, chain, file, allh):
    if type(file) == type(''):
        f = open(file, 'w')
    else:
        f = file

    fmt2 = 'ATOM  %5d %-4s %-4s %4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
    f.write('COMPND    %s\n' % title)
    seq = []
    nres = len(chain)
    line_num = 1
    for i in xrange(0, nres, 13):
        prefix = 'SEQRES%4d  %4d  ' % (line_num, nres)
        rec = [prefix]
        for j in range(i, min(i+13, nres)):
            rec.append(' '+chain[j].name)
        rec.append('\n')
        f.write(string.join(rec, ''))
        line_num = line_num + 1
    i = 1
    j = 1
    for res in chain:
        res_name = res.name
        for atom in res:
            name = atom.name
            if name == 'DU': continue
            f.write(fmt1 % ( i, name, res_name, j, atom.xyz[0], atom.xyz[1],
                            atom.xyz[2], 1.0, 0.0))
            i = i + 1
        j = j + 1
    f.write('END\n')
    f.close()

def write_redux_pdb(title, ATOMLIST,out_file):
    if type(out_file) == type(''):
        f = open(out_file, 'w')
    else:
        f = out_file
 

    f.write('REMARK    %s\n' % title)
    pdbfmt = 'ATOM  %5d  %-3s %-3s  %4d    %8.3f%8.3f%8.3f  0.00  0.00\n'
    for i in xrange(len(ATOMLIST)):
        a = ATOMLIST[i]
        f.write(pdbfmt % (a.indx,a.name,a.rnam,int(a.rnum),a.x,a.y,a.z))
    f.write('END\n')


def run(in_file, parm_file, out_file):
    title, residues = _read_input(in_file)
    db = Full_read_parm_file(parm_file)
    chain = make_chain(residues, db)
    build_peptide(chain)
    ATOMLIST = redux_from_full(parm_file,chain)
    write_redux_pdb(title, ATOMLIST, out_file)
#   write_pdb(title, chain, out_file, allh)


if __name__ == "__main__":
    import sys, getopt, os
    in_file = out_file = None

    args = getopt.getopt(sys.argv[1:], 'i:p:o:')[0]
    for opt, val in args:
        if opt == '-i':
            in_file = val
        elif opt == '-p':
            parm_file = val
        elif opt == '-o':
            out_file = val

    if in_file is None:
        in_file = sys.stdin
        ##in_file = raw_input('Enter Name of input file: ')
        ##if not in_file:
        ##    raise ValueError, 'Input file not specified'

    if out_file is None:
        out_file = sys.stdout
        ##out_file = raw_input('Enter Name of output file: ')
        ##if not out_file:
        ##    raise ValueError, 'Output file not specified'

    run(in_file, parm_file, out_file)

