"""
Calculates torsion angles alpha and tau for CA model of protein.

Outputs resname(i), resname(i+1), alpha, tau for residues 2 to N-1.
"""

from Bio.PDB import *
import math
import sys

DEGRAD = math.pi/180.0
RADDEG = 180.0/math.pi

RC_DICT = {
(  90,-165): 'Aa',
(  90,-135): 'Ab',
(  90,-105): 'Ac',
(  80,-135): 'Ab',
(  80,-105): 'Ac',
(  90, -75): 'Ad',
(  90, -45): 'Ae',
(  90, -15): 'Af',
(  90,  15): 'Ag',
(  80,  45): 'Ah',
(  80,  75): 'Ai',
(  90,  45): 'Ah',
(  90,  75): 'Ai',
(  90, 105): 'Aj',
(  90, 135): 'Ak',
(  90, 165): 'Al',
( 100,-165 ): 'Ba',
( 100,-135 ): 'Bb',
( 100,-105 ): 'Bc',
( 100, -75 ): 'Bd',
( 100, -45 ): 'Be',
( 100, -15 ): 'Bf',
( 100,  15 ): 'Bg',
( 100,  45 ): 'Bh',
( 100,  75 ): 'Bi',
( 100, 105 ): 'Bj',
( 100, 135 ): 'Bk',
( 100, 165 ): 'Bl',
( 110,-165 ): 'Ca',
( 110,-135 ): 'Cb',
( 110,-105 ): 'Cc',
( 110, -75 ): 'Cd',
( 110, -45 ): 'Ce',
( 110, -15 ): 'Cf',
( 110,  15 ): 'Cg',
( 110,  45 ): 'Ch',
( 110,  75 ): 'Ci',
( 110, 105 ): 'Cj',
( 110, 135 ): 'Ck',
( 110, 165 ): 'Cl',
( 120,-165 ): 'Da',
( 120,-135 ): 'Db',
( 120,-105 ): 'Dc',
( 120, -75 ): 'Dd',
( 120, -45 ): 'De',
( 120, -15 ): 'Df',
( 120,  15 ): 'Dg',
( 120,  45 ): 'Dh',
( 120,  75 ): 'Di',
( 120, 105 ): 'Dj',
( 120, 135 ): 'Dk',
( 120, 165 ): 'Dl',
( 130,-165 ): 'Ea',
( 130,-135 ): 'Eb',
( 130,-105 ): 'Ec',
( 130, -75 ): 'Ed',
( 130, -45 ): 'Ee',
( 130, -15 ): 'Ef',
( 130,  15 ): 'Eg',
( 130,  45 ): 'Eh',
( 130,  75 ): 'Ei',
( 130, 105 ): 'Ej',
( 130, 135 ): 'Ek',
( 130, 165 ): 'El',
( 140,-165 ): 'Fa',
( 150,-165 ): 'Fa',
( 140,-135 ): 'Fb',
( 140,-105 ): 'Fc',
( 140, -75 ): 'Fd',
( 140, -45 ): 'Fe',
( 140, -15 ): 'Ff',
( 150, -15 ): 'Ff',
( 140,  15 ): 'Fg',
( 150,  15 ): 'Fg',
( 140,  45 ): 'Fh',
( 150,  45 ): 'Fh',
( 140,  75 ): 'Fi',
( 150,  75 ): 'Fi',
( 140, 105 ): 'Fj',
( 150, 105 ): 'Fj',
( 140, 135 ): 'Fk',
( 140, 165 ): 'Fl',
( 150, 135 ): 'Fk',
( 150, 165 ): 'Fl'}

def res_rc(alf, tau):

    ialf = int(round(alf/10.0 )) * 10

    # For the special cases of slightly outside the mesostate grid
    if ialf > 140:
        ialf = 140
    if ialf < 90:
        ialf = 90

    # special case of tau equals exactly 180
    if int(tau) == 180:
        tau = 179.0
    elif int(tau) == -180:
        tau = -179.0

    itau = -15 + int(round((tau + 15)/30.0)) * 30

    try:
        return RC_DICT[(ialf,itau)]
    except:
        return 'Xx'


def getStructure(filename):
    parser = PDBParser()
    structure = parser.get_structure(filename.split('.')[0], filename)
    return structure


def writeStructure(set):
    io=PDBIO()
    io.set_structure(set)
    io.save('out.pdb')    


class CASelect(Select):
    def accept_atom(self, atom):
        if atom.get_name() == 'CA':
            return 1
        else:
            return 0

def getAll_Coords(filename):
    result =[]
    parser=PDBParser()
    structure=parser.get_structure(filename.split('.')[0], filename)
    for model in structure.get_list():
        chain = model.get_list()
        for each in chain:
            for res in each.get_list():
                for x in res.get_list():
                    result.append(x.get_coord())
    return result

def getAll_Atoms(filename):
    result =[]
    parser=PDBParser()
    structure=parser.get_structure(filename.split('.')[0], filename)
    for model in structure.get_list():
        chain = model.get_list()
        for each in chain:
            for res in each.get_list():
                for x in res.get_list():
                    result.append(x)
    return result

def getCA_Coords(filename):
    result =[]
    parser=PDBParser()
    structure=parser.get_structure(filename.split('.')[0], filename)
    for model in structure.get_list():
        chain = model.get_list()
        for each in chain:
            for res in each.get_list():
                for x in res.get_list():
                    if x.get_name() == 'CA':
                        result.append(x.get_coord())
    return result

def getCA_Atoms(filename):
    result =[]
    parser=PDBParser()
    structure=parser.get_structure(filename.split('.')[0], filename)
    for model in structure.get_list():
        chain = model.get_list()
        for each in chain:
            for res in each.get_list():
                for x in res.get_list():
                    if x.get_name() == 'CA':
                        result.append(x)
    return result

def writeCA(set):
    caio=PDBIO()
    caio.set_structure(set)
    caio.save('CA_only.pdb', CASelect())

def run_torsion(file):
    #RETURNS ALL COORDINATES OF PROTEIN X FOR TRANSFORMATION
    set = getStructure(file)
#   atom_list=Selection.unfold_entities(set, 'A')
#   writeStructure(set)
#   writeCA(set)
#   ca_coords = getCA_Coords(file)
#   print ca_coords[0][0], ca_coords[0][1], ca_coords[0][2]
#   atoms = getAll_Atoms(file)

    ca_atoms = getCA_Atoms(file)        
#   ca_dist = []
#   ca_dist.append(0.0)
#   for i in range(len(ca_atoms)-1):
#       distance = ca_atoms[i] - ca_atoms[i+1]
#       ca_dist.append(distance)
    ca_angles = []
    ca_angles.append(0.0)
    for i in range(1,len(ca_atoms)-1):
        v1 = ca_atoms[i-1].get_vector()
        v2 = ca_atoms[i].get_vector()
        v3 = ca_atoms[i+1].get_vector()
        angle = calc_angle(v1,v2,v3)
        ca_angles.append(angle)

    ca_taus = []
    ca_taus.append(0.0)
    ca_taus.append(0.0)
    for i in range(2,len(ca_atoms)-1):
        v1 = ca_atoms[i-2].get_vector()
        v2 = ca_atoms[i-1].get_vector()
        v3 = ca_atoms[i].get_vector()
        v4 = ca_atoms[i+1].get_vector()
        torangle = calc_dihedral(v1,v2,v3,v4)
        ca_taus.append(torangle)

    ca_mesos = []
    ca_mesos.append('Xx')
    ca_mesos.append('Xx')
    for i in range(2,len(ca_atoms)-1):
        alf = RADDEG * ca_angles[i]
        tau = RADDEG * ca_taus[i]
        meso =  res_rc(alf, tau)
        ca_mesos.append(meso)



    for i in range(2,len(ca_atoms)-1):
        res_nam_i = (ca_atoms[i].get_parent()).get_resname()
        res_nam_nxt = (ca_atoms[i+1].get_parent()).get_resname()
        res_id = (ca_atoms[i].get_parent()).get_id()

        sys.stdout.write('%4d %3s %3s %7.2f %7.2f  %2s\n' % (res_id[1], 
             res_nam_i, res_nam_nxt, RADDEG * (ca_angles[i]), RADDEG * (ca_taus[i]),ca_mesos[i]))
       

if __name__ == "__main__":
    filename = sys.argv[1]
    run_torsion(filename)
